import numpy as np

# reverse = True: descending order ( for example TOPSIS, CODAS), False: ascending order (for example VIKOR, SPOTIS)
def rank_preferences(pref, reverse = True):
    """Rank alternatives according to MCDA preference function values. If more than one alternative
    have the same preference function value, they will be given the same rank value (tie).
    Parameters
    ----------
        pref : ndarray
            vector with MCDA preference function values for alternatives
        reverse : bool
            Boolean variable which is True for MCDA methods which rank alternatives in descending
            order and False for MCDA methods which rank alternatives in ascending
            order
    Returns
    -------
        ndarray
            vector with alternatives ranking. Alternative with 1 value is the best.
    """

    # Create an array ndarray for the ranking values of alternatives
    rank = np.zeros(len(pref))
    # Generate sorted vector with MCDA preference function values
    # sorting order is determined by variable `reverse` and depends on MCDA method
    sorted_pref = sorted(pref, reverse = reverse)
    # position of the best alternative in ranking is denoted by 1, so assign 1 to `pos`
    pos = 1
    for i in range(len(sorted_pref) - 1):
        # find index in vector with preference values `pref` equal to preference value in sorted vector `sorted_pref`
        ind = np.where(sorted_pref[i] == pref)[0]
        # assign rank denoted by `pos` to index `ind` in rank vector
        rank[ind] = pos
        # if the next preference value in sorted vector is higher than actual, increase `pos` which denotes rank
        # if the next preference value is equal to actual, `pos` will be unchanged
        if sorted_pref[i] != sorted_pref[i + 1]:
            pos += 1
    # find index with the last preference value
    ind = np.where(sorted_pref[i + 1] == pref)[0]
    # assign the last place `pos` to alternative with the last preference value
    rank[ind] = pos
    return rank.astype(int)


import numpy as np
import copy
import itertools


# euclidean distance
def euclidean(A, B):
    """
    Calculate Euclidean distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    tmp = np.sqrt(np.sum(np.square(A - B)))
    return tmp

# manhattan distance
def manhattan(A, B):
    """
    Calculate Manhattan (Taxicab) distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    tmp = np.sum(np.abs(A - B))
    return tmp


# for hausdorff distance
def hausdorff_distance(A, B):
    min_h = np.inf
    for i, j in itertools.product(range(len(A)), range(len(B))):
        d = euclidean(A[i], B[j])
        if d < min_h:
            min_h = d
            min_ind = j

    max_h = -np.inf
    for i in range(len(A)):
        d = euclidean(A[i], B[min_ind])
        if d > max_h:
            max_h = d

    return max_h


# hausdorff distance
def hausdorff(A, B):
    """
    Calculate Hausdorff distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    ah = hausdorff_distance(A, B)
    bh = hausdorff_distance(B, A)
    tmp = max(ah, bh)
    return tmp


# correlation distance
def correlation(A, B):
    """
    Calculate Correlation distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    numerator = np.sum((A - np.mean(A)) * (B - np.mean(B)))
    denominator = np.sqrt(np.sum((A - np.mean(A)) ** 2)) * np.sqrt(np.sum((B - np.mean(B)) ** 2))
    if denominator == 0:
        denominator = 1
    tmp = 1 - (numerator / denominator)
    return tmp


# chebyshev distance
def chebyshev(A, B):
    """
    Calculate Chebyshev distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    max_h = -np.inf
    for i, j in itertools.product(range(len(A)), range(len(B))):
        d = np.abs(A[i] - B[j])
        if d > max_h:
            max_h = d

    return max_h


# standardized euclidean distance
def std_euclidean(A, B):
    """
    Calculate Standardized Euclidean distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    tab_std = np.vstack((A, B))
    stdv = np.sum(np.square(tab_std - np.mean(tab_std, axis = 0)), axis = 0)
    stdv = np.sqrt(stdv / tab_std.shape[0])
    stdv[stdv == 0] = 1
    tmp = np.sum(np.square((A - B) / stdv))
    tmp = np.sqrt(tmp)
    return tmp


# cosine distance
def cosine(A, B):
    """
    Calculate Cosine distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    numerator = np.sum(A * B)
    denominator = (np.sqrt(np.sum(np.square(A)))) * (np.sqrt(np.sum(np.square(B))))
    if denominator == 0:
        denominator = 1
    tmp = 1 - (numerator / denominator)
    return tmp


# cosine similarity measure
def csm(A, B):
    """
    Calculate Cosine similarity measure of distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    numerator = np.sum(A * B)
    denominator = (np.sqrt(np.sum(A))) * (np.sqrt(np.sum(B)))
    if denominator == 0:
        denominator = 1
    tmp = numerator / denominator
    return tmp


# squared euclidean distance
def squared_euclidean(A, B):
    """
    Calculate Squared Euclidean distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    tmp = np.sum(np.square(A - B))
    return tmp


# sorensen or bray-curtis distance
def bray_curtis(A, B):
    """
    Calculate Bray-Curtis distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    numerator = np.sum(np.abs(A - B))
    denominator = np.sum(A + B)
    if denominator == 0:
        denominator = 1
    tmp = numerator / denominator
    return tmp


# canberra distance
def canberra(A, B):
    """
    Calculate Canberra distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    numerator = np.abs(A - B)
    denominator = A + B
    denominator[denominator == 0] = 1
    tmp = np.sum(numerator / denominator)
    return tmp


# lorentzian distance
def lorentzian(A, B):
    """
    Calculate Lorentzian distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    tmp = np.sum(np.log(1 + np.abs(A - B)))
    return tmp


# jaccard distance
def jaccard(A, B):
    """
    Calculate Jaccard distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    numerator = np.sum(np.square(A - B))
    denominator = np.sum(A ** 2) + np.sum(B ** 2) - np.sum(A * B)
    if denominator == 0:
        denominator = 1
    tmp = numerator / denominator
    return tmp


# dice distance
def dice(A, B):
    """
    Calculate Dice distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    numerator = np.sum(np.square(A - B))
    denominator = np.sum(A ** 2) + np.sum(B ** 2)
    if denominator == 0:
        denominator = 1
    tmp = numerator / denominator
    return tmp


# bhattacharyya distance
def bhattacharyya(A, B):
    """
    Calculate Bhattacharyya distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    value = (np.sum(np.sqrt(A * B)))**2
    if value == 0:
        tmp = 0
    else:
        tmp = -np.log(value)
    return tmp


# hellinger distance
def hellinger(A, B):
    """
    Calculate Hellinger distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    value = 1 - np.sum(np.sqrt(A * B))
    if value < 0:
        value = 0
    tmp = 2 * np.sqrt(value)
    return tmp


# matusita distance
def matusita(A, B):
    """
    Calculate Matusita distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    value = 2 - 2 * (np.sum(np.sqrt(A * B)))
    if value < 0:
        value = 0
    tmp = np.sqrt(value)
    return tmp


# squared-chord distance
def squared_chord(A, B):
    """
    Calculate Squared-Chord distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    tmp = np.sum(np.square(np.sqrt(A) - np.sqrt(B)))
    return tmp


# pearson chi-square distance
def pearson_chi_square(A, B):
    """
    Calculate Pearson Chi Square distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    numerator = np.square(A - B)
    denominator = copy.deepcopy(B)
    denominator[denominator == 0] = 1
    tmp = np.sum(numerator / denominator)
    return tmp

# squared chi-square distance
def squared_chi_square(A, B):
    """
    Calculate Squared Chi Sqaure distance between two vectors `A` and `B`.

    Parameters
    ----------
        A : ndarray
            First vector containing values
        B : ndarray
            Second vector containing values

    Returns
    -------
        float
            distance value between two vetors
    """
    numerator = np.square(A - B)
    denominator = A + B
    denominator[denominator == 0] = 1
    tmp = np.sum(numerator / denominator)
    return tmp


from abc import ABC
import numpy as np

class MCDA_method(ABC):

    def __call__(self, matrix, weights, types):
        """Score alternatives from decision matrix `matrix` using criteria weights `weights` and criteria types `types`

        Parameters
        ----------
            matrix : ndarray
                decision matrix with performance values for m alternatives in rows and n criteria in columns
            weights : ndarray
                vector with criteria weights with size equal to number of columns n of `matrix`
            types : ndarray
                vector with criteria types containing values of 1 for profit criteria and -1 for cost criteria
                with size equal to number of columns n of `matrix`
        """
        pass

    @staticmethod
    def _verify_input_data(matrix, weights, types):
        m, n = matrix.shape
        if len(weights) != n:
            raise ValueError('The size of the weight vector must be the same as the number of criteria')
        if len(types) != n:
            raise ValueError('The size of the types vector must be the same as the number of criteria')
        check_types = np.all((types == 1) | (types == -1))
        if check_types == False:
            raise ValueError('Criteria types can only have a value of 1 for profits and -1 for costs')


from . import mcda_method
from .codas import CODAS
from .edas import EDAS
from .mabac import MABAC
from .multimoora import MULTIMOORA
from .spotis import SPOTIS
from .topsis import TOPSIS
from .vikor import VIKOR
from .waspas import WASPAS


import numpy as np
import itertools

from .correlations import pearson_coeff
from .normalizations import sum_normalization, minmax_normalization


# equal weighting
def equal_weighting(matrix):
    """
    Calculate criteria weights using objective Equal weighting method
    Parameters
    ----------
        matrix : ndarray
            Decision matrix with performance values of m alternatives and n criteria

    Returns
    -------
        ndarray
            vector of criteria weights
    """
    N = np.shape(matrix)[1]
    w = np.ones(N) / N
    return w


# Entropy weighting
def entropy_weighting(matrix):
    """
    Calculate criteria weights using objective Entropy weighting method
    Parameters
    ----------
        matrix : ndarray
            Decision matrix with performance values of m alternatives and n criteria

    Returns
    -------
        ndarray
            vector of criteria weights
    """
    # normalize the decision matrix with sum_normalization method from normalizations as for profit criteria
    types = np.ones(np.shape(matrix)[1])
    pij = sum_normalization(matrix, types)
    # Transform negative values in decision matrix `matrix` to positive values
    pij = np.abs(pij)
    m, n = np.shape(pij)
    H = np.zeros((m, n))

    # Calculate entropy
    for j, i in itertools.product(range(n), range(m)):
        if pij[i, j]:
            H[i, j] = pij[i, j] * np.log(pij[i, j])

    h = np.sum(H, axis = 0) * (-1 * ((np.log(m)) ** (-1)))

    # Calculate degree of diversification
    d = 1 - h

    # Set w as the degree of importance of each criterion
    w = d / (np.sum(d))
    return w


# Standard Deviation weighting
def std_weighting(matrix):
    """
    Calculate criteria weights using objective Standard deviation weighting method
    Parameters
    ----------
        matrix : ndarray
            Decision matrix with performance values of m alternatives and n criteria
            
    Returns
    -------
        ndarray
            vector of criteria weights
    """

    # Calculate the standard deviation of each criterion in decision matrix
    stdv = np.sqrt((np.sum(np.square(matrix - np.mean(matrix, axis = 0)), axis = 0)) / (matrix.shape[0]))
    # Calculate criteria weights by dividing the standard deviations by their sum
    w = stdv / np.sum(stdv)
    return w


# CRITIC weighting
def critic_weighting(matrix):
    """
    Calculate criteria weights using objective CRITIC weighting method
    Parameters
    ----------
        matrix : ndarray
            Decision matrix with performance values of m alternatives and n criteria
            
    Returns
    -------
        ndarray
            vector of criteria weights
    """
    # Normalize the decision matrix using Minimum-Maximum normalization minmax_normalization from normalizations as for profit criteria
    types = np.ones(np.shape(matrix)[1])
    x_norm = minmax_normalization(matrix, types)
    # Calculate the standard deviation
    std = np.std(x_norm, axis = 0)
    n = np.shape(x_norm)[1]
    # Calculate correlation coefficients of all pairs of columns of normalized decision matrix
    correlations = np.zeros((n, n))
    for i, j in itertools.product(range(n), range(n)):
        correlations[i, j] = pearson_coeff(x_norm[:, i], x_norm[:, j])

    # Calculate the difference between 1 and calculated correlations
    difference = 1 - correlations
    # Multiply the difference by the standard deviation
    C = std * np.sum(difference, axis = 0)
    # Calculate the weights by dividing vector with C by their sum
    w = C / np.sum(C)
    return w